@@ -5,7 +5,7 @@ use Carp qw/carp croak confess/;
 
 BEGIN {
   $Math::Prime::Util::PP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::PP::VERSION = '0.27';
+  $Math::Prime::Util::PP::VERSION = '0.29';
 }
 
 # The Pure Perl versions of all the Math::Prime::Util routines.
@@ -43,7 +43,17 @@ our $_Infinity = 0+'inf';
 $_Infinity = 20**20**20 if 65535 > $_Infinity;   # E.g. Windows
 our $_Neg_Infinity = -$_Infinity;
 
-my $_have_MPFR = -1;
+{
+  my $_have_MPFR = -1;
+  sub _MPFR_available {
+    if ($_have_MPFR < 0) {
+      $_have_MPFR = 0;
+      $_have_MPFR = 1 if (!defined $ENV{MPU_NO_MPFR} || $ENV{MPU_NO_MPFR} != 1)
+                      && eval { require Math::MPFR; $Math::MPFR::VERSION>=2.03; };
+    }
+    return $_have_MPFR;
+  }
+}
 
 my $_precalc_size = 0;
 sub prime_precalc {
@@ -83,13 +93,13 @@ sub _validate_positive_integer {
   croak "Parameter '$n' must be <= $max" if defined $max && $n > $max;
   if (ref($_[0])) {
     $_[0] = Math::BigInt->new("$_[0]") unless ref($_[0]) eq 'Math::BigInt';
-    # Stupid workaround for Math::BigInt::GMP RT # 71548
+    # Workarounds for Math::BigInt::GMP RT # 71548 and Perl 5.6
     if ($_[0]->bacmp(''.~0) <= 0) {
-      $_[0] = int($_[0]->bstr);
-      croak "Didn't convert!" if ref($_[0]);
-    } else {
-      $_[0]->upgrade(undef) if $_[0]->upgrade();  # Stop BigFloat upgrade
+      my $intn = int($_[0]->bstr);
+      $_[0] = $intn if $_[0]->bcmp("$intn") == 0;
     }
+    # Stop BigFloat upgrade
+    $_[0]->upgrade(undef) if ref($_[0]) && $_[0]->upgrade();
   } else {
     if ($n > ~0) {
       croak "Parameter '$n' outside of integer range" if !defined $bigint::VERSION;
@@ -748,6 +758,24 @@ sub _log2 {
 
 
 
+sub is_pseudoprime {
+  my($n, $base) = @_;
+  _validate_positive_integer($n);
+  _validate_positive_integer($base);
+
+  return 1 if $n == 2 || $n == 3;
+  return 0 if $n < 5;
+  croak "Base $base is invalid" if $base < 2;
+  if ($base >= $n) {
+    $base = $base % $n;
+    return 1 if $base <= 1 || $base == $n-1;
+  }
+  my $x = (ref($n) eq 'Math::BigInt')
+        ? $n->copy->bzero->badd($base)->bmodpow($n-1,$n)
+        : _native_powmod($base, $n-1, $n);
+  return ($x == 1);
+}
+
 sub miller_rabin {
   my($n, @bases) = @_;
   _validate_positive_integer($n);
@@ -821,6 +849,26 @@ sub miller_rabin {
   1;
 }
 
+sub _is_perfect_square {
+  my($n) = @_;
+
+  if (ref($n) eq 'Math::BigInt') {
+    my $mc = int(($n & 31)->bstr);
+    if ($mc==0||$mc==1||$mc==4||$mc==9||$mc==16||$mc==17||$mc==25) {
+      my $sq = $n->copy->bsqrt->bfloor;
+      $sq->bmul($sq);
+      return 1 if $sq == $n;
+    }
+  } else {
+    my $mc = $n & 31;
+    if ($mc==0||$mc==1||$mc==4||$mc==9||$mc==16||$mc==17||$mc==25) {
+      my $sq = int(sqrt($n));
+      return 1 if ($sq*$sq) == $n;
+    }
+  }
+  0;
+}
+
 # Calculate Jacobi symbol (M|N)
 sub _jacobi {
   my($n, $m) = @_;
@@ -875,42 +923,18 @@ sub _find_jacobi_d_sequence {
   return ($sign * $d);
 }
 
+sub is_lucas_pseudoprime {
+  return is_strong_lucas_pseudoprime($_[0], 'weak');
+}
 
 sub is_strong_lucas_pseudoprime {
-  my($n) = @_;
+  my($n, $doweak) = @_;
   _validate_positive_integer($n);
-
-  # We're trying to limit the bignum calculations as much as possible.
-  # It's also important to try to match whatever they passed in.  For instance
-  # if they use a GMP or Pari object, we must do the same.  Hence instead of:
-  #        my $U = Math::BigInt->bone;
-  # we do
-  #        my $U = $n->copy->bone;
-  # so U is the same class as n.  If they passed in a string or a small value,
-  # then we just make it up.
+  $doweak = (defined $doweak && $doweak eq 'weak') ? 1 : 0;
 
   return 1 if $n == 2;
   return 0 if $n < 2 || ($n % 2) == 0;
-
-  # References:
-  #     http://www.trnicely.net/misc/bpsw.html
-  #     Math::Primality
-
-  # Check for perfect square
-  if (ref($n) eq 'Math::BigInt') {
-    my $mc = int(($n & 31)->bstr);
-    if ($mc==0||$mc==1||$mc==4||$mc==9||$mc==16||$mc==17||$mc==25) {
-      my $sq = $n->copy->bsqrt->bfloor;
-      $sq->bmul($sq);
-      return 0 if $sq == $n;
-    }
-  } else {
-    my $mc = $n & 31;
-    if ($mc==0||$mc==1||$mc==4||$mc==9||$mc==16||$mc==17||$mc==25) {
-      my $sq = int(sqrt($n));
-      return 0 if ($sq*$sq) == $n;
-    }
-  }
+  return 0 if _is_perfect_square($n);
 
   # Determine Selfridge D, P, and Q parameters
   my $D = _find_jacobi_d_sequence($n);
@@ -936,15 +960,18 @@ sub is_strong_lucas_pseudoprime {
   #   my $d=$m->copy; my $s=0; while ($d->is_even) { $s++; $d->brsft(1); }
   #   die "Invalid $m, $d, $s\n" unless $m == $d * 2**$s;
   my $dstr = substr($m->as_bin, 2);
-  $dstr =~ s/(0*)$//;
-  my $s = length($1);
+  my $s = 0;
+  if (!$doweak) {
+    $dstr =~ s/(0*)$//;
+    $s = length($1);
+  }
 
   my $ZERO = $n->copy->bzero;
   my $U = $ZERO + 1;
   my $V = $ZERO + $P;
   my $Qk = $ZERO + $Q;
   my $bpos = 0;
-  while ($bpos++ < length($dstr)) {
+  while (++$bpos < length($dstr)) {
     $U = ($U * $V) % $n;
     $V = ($V * $V - 2*$Qk) % $n;
     $Qk = ($Qk * $Qk) % $n;
@@ -963,6 +990,7 @@ sub is_strong_lucas_pseudoprime {
       $Qk = ($Qk * $Q) % $n;
     }
   }
+  if ($doweak) { return $U->is_zero ? 1 : 0; }
   return 1 if $U->is_zero || $V->is_zero;
 
   # Compute powers of V
@@ -974,6 +1002,71 @@ sub is_strong_lucas_pseudoprime {
   return 0;
 }
 
+sub is_extra_strong_lucas_pseudoprime {
+  my($n) = @_;
+  _validate_positive_integer($n);
+
+  return 1 if $n == 2;
+  return 0 if $n < 2 || ($n % 2) == 0;
+  return 0 if _is_perfect_square($n);
+
+  my ($P, $Q, $D) = (3, 1, 5);
+  while (1) {
+    my $gcd = (ref($n) eq 'Math::BigInt') ? Math::BigInt::bgcd($D, $n)
+                                          : _gcd_ui($D, $n);
+    return 0 if $gcd > 1 && $gcd != $n;  # Found divisor $d
+    last if _jacobi($D, $n) == -1;
+    $P++;
+    $D = $P*$P - 4;
+  }
+  die "Lucas incorrect DPQ: $D, $P, $Q\n" if ($D != $P*$P - 4*$Q);
+
+  if (ref($n) ne 'Math::BigInt') {
+    if (!defined $Math::BigInt::VERSION) {
+      eval { require Math::BigInt;  Math::BigInt->import(try=>'GMP,Pari'); 1; }
+      or do { croak "Cannot load Math::BigInt "; }
+    }
+    $n = Math::BigInt->new("$n");
+  }
+
+  my $m = $n->copy->badd(1);
+  # Traditional d,s:
+  #   my $d=$m->copy; my $s=0; while ($d->is_even) { $s++; $d->brsft(1); }
+  #   die "Invalid $m, $d, $s\n" unless $m == $d * 2**$s;
+  my $dstr = substr($m->as_bin, 2);
+  $dstr =~ s/(0*)$//;
+  my $s = length($1);
+
+  my $ZERO = $n->copy->bzero;
+  my $U = $ZERO + 1;
+  my $V = $ZERO + $P;
+  my $bpos = 0;
+  while (++$bpos < length($dstr)) {
+    $U->bmul($V)->bmod($n);
+    $V->bmul($V)->bsub(2)->bmod($n);
+    if (substr($dstr,$bpos,1)) {
+      my $U_times_D = $U->copy->bmul($D);
+      $U->bmul($P)->badd($V);
+      $U->badd($n) if $U->is_odd;
+      $U->brsft(1);
+
+      $V->bmul($P)->badd($U_times_D);
+      $V->badd($n) if $V->is_odd;
+      $V->brsft(1);
+    }
+  }
+  $U->bmod($n);
+  $V->bmod($n);
+  return 1 if $U->is_zero && ($V == 2 || $V == ($n-2));
+  return 1 if $V->is_zero;
+
+  foreach my $r (1 .. $s-1) {
+    $V->bmul($V)->bsub(2)->bmod($n);
+    return 1 if $V->is_zero;
+  }
+  return 0;
+}
+
 
 my $_poly_bignum;
 sub _poly_new {
@@ -1485,7 +1578,7 @@ sub pminus1_factor {
   while (1) {
     $pc_end = $B1 if $pc_end > $B1;
     @bprimes = @{ primes($pc_beg, $pc_end) };
-    foreach my $q (@bprimes) { 
+    foreach my $q (@bprimes) {
       my($k, $kmin) = ($q, int($B1 / $q));
       while ($k <= $kmin) { $k *= $q; }
       $t *= $k;                         # accumulate powers for a
@@ -2038,14 +2131,8 @@ sub ExponentialIntegral {
   return 0              if $x == $_Neg_Infinity;
   return $_Infinity     if $x == $_Infinity;
 
-  # Use MPFR if possible.
-  if ($_have_MPFR < 0) {
-    $_have_MPFR = 0;
-    $_have_MPFR = 1 if (!defined $ENV{MPU_NO_MPFR} || $ENV{MPU_NO_MPFR} != 1)
-                    && eval { require Math::MPFR; $Math::MPFR::VERSION>=2.03; };
-  }
   # Gotcha -- MPFR decided to make negative inputs return NaN.  Grrr.
-  if ($_have_MPFR && $x > 0) {
+  if ($x > 0 && _MPFR_available()) {
     my $wantbf = 0;
     my $xdigits = 17;
     if (defined $bignum::VERSION || ref($x) =~ /^Math::Big/) {
@@ -2148,14 +2235,8 @@ sub LogarithmicIntegral {
   return $_Infinity     if $x == $_Infinity;
   croak "Invalid input to LogarithmicIntegral:  x must be > 0" if $x <= 0;
 
-  # Use MPFR if possible.
-  if ($_have_MPFR < 0) {
-    $_have_MPFR = 0;
-    $_have_MPFR = 1 if (!defined $ENV{MPU_NO_MPFR} || $ENV{MPU_NO_MPFR} != 1)
-                    && eval { require Math::MPFR; $Math::MPFR::VERSION>=2.03; };
-  }
   # Remember MPFR eint doesn't handle negative inputs
-  if ($_have_MPFR && $x >= 1) {
+  if ($x >= 1 && _MPFR_available()) {
     my $wantbf = 0;
     my $xdigits = 17;
     if (defined $bignum::VERSION || ref($x) =~ /^Math::Big/) {
@@ -2282,12 +2363,7 @@ sub RiemannZeta {
   my($x) = @_;
 
   # Use MPFR if possible.
-  if ($_have_MPFR < 0) {
-    $_have_MPFR = 0;
-    $_have_MPFR = 1 if (!defined $ENV{MPU_NO_MPFR} || $ENV{MPU_NO_MPFR} != 1)
-                    && eval { require Math::MPFR; $Math::MPFR::VERSION>=2.03; };
-  }
-  if ($_have_MPFR) {
+  if (_MPFR_available()) {
     my $wantbf = 0;
     my $xdigits = 17;
     if (defined $bignum::VERSION || ref($x) =~ /^Math::Big/) {
@@ -2373,12 +2449,7 @@ sub RiemannR {
   croak "Invalid input to ReimannR:  x must be > 0" if $x <= 0;
 
   # Use MPFR if possible.
-  if ($_have_MPFR < 0) {
-    $_have_MPFR = 0;
-    $_have_MPFR = 1 if (!defined $ENV{MPU_NO_MPFR} || $ENV{MPU_NO_MPFR} != 1)
-                    && eval { require Math::MPFR; $Math::MPFR::VERSION>=2.03; };
-  }
-  if ($_have_MPFR) {
+  if (_MPFR_available()) {
     my $wantbf = 0;
     my $xdigits = 17;
     if (defined $bignum::VERSION || ref($x) =~ /^Math::Big/) {
@@ -2478,7 +2549,7 @@ Math::Prime::Util::PP - Pure Perl version of Math::Prime::Util
 
 =head1 VERSION
 
-Version 0.27
+Version 0.29
 
 
 =head1 SYNOPSIS
@@ -2653,6 +2724,13 @@ methods take quite a bit of time and space.  Think abut whether a bound or
 approximation would be acceptable instead.
 
 
+=head2 is_pseudoprime
+
+Takes a positive number C<n> and a base C<a> as input, and returns 1 if
+C<n> is a probable prime to base C<a>.  This is the simple Fermat primality
+test.  Removing primes, given base 2 this produces the sequence
+L<OEIS A001567|http://oeis.org/A001567>.
+
 =head2 is_strong_pseudoprime
 
 =head2 miller_rabin
@@ -2661,18 +2739,25 @@ approximation would be acceptable instead.
   my $probably_prime = is_strong_pseudoprime($n, 2, 3, 5, 7, 11, 13, 17);
 
 Takes a positive number as input and one or more bases.  The bases must be
-greater than 1.  Returns 2 is C<n> is definitely prime, 1 if C<n>
-is probably prime, and 0 if C<n> is definitely composite.  Since this is
-just the Miller-Rabin test, a value of 2 is only returned for inputs of
-2 and 3, which are shortcut.  If 0 is returned, then the number really is a
-composite.  If 1 is returned, there is a good chance the number is prime
-(depending on the input and the bases), but we cannot be sure.
+greater than C<1>.  Returns 1 if the input is a strong probable prime
+to all of the bases, and 0 if not.
+
+If 0 is returned, then the number really is a composite.  If 1 is returned,
+then it is either a prime or a strong pseudoprime to all the given bases.
+Given enough distinct bases, the chances become very, very strong that the
+number is actually prime.
 
 This is usually used in combination with other tests to make either stronger
 tests (e.g. the strong BPSW test) or deterministic results for numbers less
-than some verified limit (such as the C<is_prob_prime> function in this module).
+than some verified limit.
 
 
+=head2 is_lucas_pseudoprime
+
+Takes a positive number as input, and returns 1 if the input is a standard
+Lucas probable prime using the Selfridge method of choosing D, P, and Q (some
+sources call this a Lucas-Selfridge pseudoprime).
+
 =head2 is_strong_lucas_pseudoprime
 
 Takes a positive number as input, and returns 1 if the input is a strong
@@ -2681,6 +2766,14 @@ sources call this a strong Lucas-Selfridge pseudoprime).  This is one half
 of the BPSW primality test (the Miller-Rabin strong pseudoprime test with
 base 2 being the other half).
 
+=head2 is_extra_strong_lucas_pseudoprime
+
+Takes a positive number as input, and returns 1 if the input passes the extra
+strong Lucas test (as defined in
+L<Grantham 2000|http://www.ams.org/mathscinet-getitem?mr=1680879>).
+This has slightly more restrictive conditions than the strong Lucas test,
+but uses different starting parameters so is not directly comparable.
+
 =head2 is_aks_prime
 
 Takes a positive number as input, and returns 1 if the input can be proven
