@@ -6,6 +6,9 @@
 
 extern unsigned char* sieve_erat30(UV end);
 extern int sieve_segment(unsigned char* mem, UV startd, UV endd);
+extern void* start_segment_primes(UV low, UV high, unsigned char** segmentmem);
+extern int next_segment_primes(void* vctx, UV* base, UV* low, UV* high);
+extern void end_segment_primes(void* vctx);
 
 
 static const UV wheel30[] = {1, 7, 11, 13, 17, 19, 23, 29};
